# Relative_RGB_color_mask_maker 8 & 16bit Version
# Ver. 0.25e (2022.8)
# add tanceparency factor
# add luminosity mask (0.12)
# add Monochrome film support (0.13)
# add imp2 mixing ratio parameter (0.14) for preparing next version
# add Mask Inverting Option for Exclude Selected Area (0.15)
# add function of making combined mask (0.20)
# bugfix (0.22)
# User Interface Renewal (0.25)
# add Luminosity mask preview (0.25c)
# improve Luminosity mask preview (0.25d)
# minor modify in  interface (0.25e)
# (c)2021-22 Ohnishi, Yasuo
from ij import IJ, ImagePlus
from ij import WindowManager as WM
from ij.gui import GenericDialog, ImagePanel
from ij.io import DirectoryChooser, OpenDialog
from ij import plugin
from ij.plugin import ChannelSplitter, Duplicator
from ij.plugin import RGBStackMerge, RGBStackConverter
from ij.plugin import Duplicator
from ij.plugin.filter import ThresholdToSelection
from ij.process import ImageConverter, ImageProcessor
from ij.process import ByteProcessor
from loci.plugins import BF
from java.awt.event import AdjustmentListener, ItemListener, TextListener
import os
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')

#---Convert Value in 8, 16, 32bit mode
def convValue(Value, Mode):
#     Converting Luminance Value each in 8, 16, 32bit mode
        if Mode == "16-bit":
                Value = Value * 256
        elif Mode == "32-bit":
                Value = Value * 256 * 256
        else:
                Value = Value
        return Value

#--------------------------------------------
def grayscale_mask_maker(imp1, imp2, imp3, YTh, Mode, blInvert, Show, colorCh,\
    TrFac, Gaussian, GaussianSigma, imp2MixRatio):
# this is making relative color mask routine
    """
    parameters
        imp1: Target Channel (B), imp2: Comparative Channel (G)
        imp3: Comparative Channnel 2 (R)
        YTh: Target Color Ch Threshold (originally Yellow Threshold)
        blInvert: Color invert Yes or No (originally invert for blue mask) 
        TrFac: Transparency Factor
        Gaussian: Apply or not Gaussian Blur
        GaussianSigma: Gaussian Blur Sigma
        imp2MixRatio: imp2 Mixing Ratio (to imp3) 0.0 to 1.0 Normally fixed to 0.50
    """
# Comparative (G) ch.1 x 0.5
    imp2_05 = Duplicator().run(imp2)
    Vstr = str(imp2MixRatio)  # imp2 mixing ratio
    IJ.run(imp2_05, "Multiply...", "value=" + Vstr)
    if Show == "Yes":
        imp2_05.show()
# Comparative (R) ch.2 ch. x 0.5
    imp3_05 = Duplicator().run(imp3) # imp3 mixing ratio
    Vstr = str(1.0 - imp2MixRatio)
    IJ.run(imp3_05, "Multiply...", "value=" + Vstr)
    if Show == "Yes":
        imp3_05.show()
# Comparative (G) Ch. 1 x 0.5 + (R) Ch.2 x 0.5
    ic = plugin.ImageCalculator()
    imp32 = ic.run("Add created", imp3_05, imp2_05)
# Close Windows
    if Show == "Yes":
        imp3.hide()
        imp2.hide()
        imp1.hide()
        imp32.show()
        imp2_05.hide()
        imp3_05.hide()
    imp2_05.close()
    imp3_05.close()
# (Ch1(R) x 0.5 + Ch2(G) x 0.5) - Target Ch (B)
    ic = plugin.ImageCalculator()
# Make Target Color (Yellow) Image Files 
# in each Threshold, Make 3 kinds of Image Files: x1, x2, x4 
    V1 = YTh
# Target (Yellow) Mask Threshold
    if blInvert == True: # for Blue mask
        YTh = YTh * -1 # Invert the threshold value for Blue mask
    V2 = str(convValue(YTh, Mode)) # actual Yellow mask Threshold
        # for making gray scale mask
           # Set the Threshold in R + G Image
    IJ.run(imp32, "Add...","value=" + V2)
    V = str(abs(V1))
    if V1 >= 0:
        V= "+" + V
    else:
        V= "-" + V
    if blInvert == False: # for making Yellow mask
        impGraymask = ic.run("Subtract created", imp32, imp1) 
                                            # Make Yellow Mask
        if colorCh == "Yellow/Blue":
            V = "Y" + V
        elif colorCh == "Cyan/Red":
            V = "Cy" + V
        elif colorCh == "Magenta/Green":
            V = "Mg" + V
    else:                 # for making Blue mask
        impGraymask = ic.run("Subtract created", imp1, imp32) 
                                            # Make Blue Mask
        if colorCh == "Yellow/Blue":
            V = "B" + V
        elif colorCh == "Cyan/Red":
            V = "R" + V
        elif colorCh == "Magenta/Green":
            V = "G" + V

    if Show == "Yes":
        impGraymask.show()  # Show the subtracted image in window
    if TrFac < 100: # making grayscale mask
        IJ.run(impGraymask, "Multiply...", "value=" + str(4 * TrFac)) 
            # making mask more transparent (x 4)
    else: # if TrFac = 100, making relative color mask binary (calling subroutine)
        Gmaskimp = make_binary_mask(impGraymask, 1, convValue(255, Mode), Mode, Show)
        if Show == "Yes":
            impGraymask.hide()  # Hide the subtracted image in window
        impGraymask = Gmaskimp
#   apply Gaussian blur
#    if Gaussian == True:
#        IJ.run(impGraymask, "Gaussian Blur...", "sigma=" + str(GaussianSigma))
#   closing working windows
    imp32.hide()
    imp32.close()
    return impGraymask, V
    """
    return values
        impGraymask: mask image
        V: description for mask parameters
    """
#-------------------------------------------
def luminosity_mask_maker(imp1, LTh, Mode, NegaInvert, Show, TrFac):
    """
    parameters
        imp1: Grayscale image
        LTh: Zone lower Threshold
        NegaInvert: Nega invert True or False
        TrFac: Transparency Factor
    """
    if Show == "Yes":
        imp1.show()
# Make Grayscale Mask Image Files 
    V1 = str(LTh) # for threshold expression in file name
# Target (Yellow) Mask Threshold
    V2 = str(convValue(abs(LTh), Mode)) # actual Yellow mask Threshold
    IJ.run(imp1, "Subtract...","value=" + V2)
    if NegaInvert == True: # for making inverted Negative mask
        IJ.run(imp1, "Multiply...", "value=" + str(4 / TrFac)) 
            # making mask more transparent (x 4 for Nega)
        imp1.getProcessor().invert()
        V1 = V1 + "_Nega"
        V = "_Nega_x "+ str(TrFac)
    else: # for making positive mask
        IJ.run(imp1, "Multiply...", "value=" + str(4 * TrFac)) 
            # making mask more transparent (x 4)
        V1 = V1 + "_Posi"
        V = "_Posi_x" + str(TrFac)
    impGraymask = imp1
    if Show == "Yes":
        impGraymask.show()  # Show the subtracted image in window
    return impGraymask, V, V1
    """
    return values
        impGraymask: mask image
        V: description for mask parameters
        V1: description for mask parameters
    """
#-------------------------------------------
def make_binary_mask(impZone, zoneLTh, zoneHTh, Mode, Show):
    """
    parameters
       impZone: Original Ch.image
       zoneLTh: Lower threshold, zoneHTh: Upper threshold
       Mode: bit depth, Show: Show windows

    """
# make impZone for Grayscale 8 bit!!
    ImageConverter(impZone).setDoScaling(True)
#           it is needs for 16bit image to down convert to 8bit
    ImageConverter(impZone).convertToGray8()
    ip = impZone.getProcessor()
    mask = ByteProcessor(ip.width, ip.height)
    mask.setValue(255)
    if Show == "Yes":
        impZone.show()
# make Zone mask
    impZone.getProcessor().setThreshold(zoneLTh, zoneHTh, \
                                                  ImageProcessor.NO_LUT_UPDATE)
    print "Zone L & H", zoneLTh, zoneHTh
    roi = ThresholdToSelection.run(impZone)
    mask.setRoi(roi)
    mask.fill(roi)
    try:
        maskimp = ImagePlus("Mask", mask) 
#        only 8bit Zone mask made from ROI: Zone mask: maskimp
    except: # exception process for ROI missing
        IJ.error("Selected ROI (Region of Interest) is missing.")
        if Show == "Yes":
            impZone.hide()
        impZone.close()
        return
    if Show == "Yes":
        maskimp.show() # show Zone mask

# deepen Zone Mask bit depth  (8bit to 16 or 32 bit)      
    if Mode == "16-bit":
        ImageConverter(maskimp).convertToGray16()
    elif Mode == "32-bit":
        ImageConverter(maskimp).convertToGray32()
    IJ.run(maskimp, "Multiply...", "value=" + str(convValue(1, Mode)))
#   after deepen bit depth, Save Zone Mask
    return maskimp
    """
    return values
        maskimp: mask image
    """
#----------------------------------------------
### RGB Manipulation 
def RGBmanipulation(imp0, folder,filename, baseCh, parm, colorCh, stackSize,\
                                                                      sequence):
# this is mask making & saving main routine
    """
    parameters
        imp0: original image
        folder: folder name of original image file
        filename: file name of original image file
        baseCh: channel name for deciding mask upper & lower thresholds
        parm: parameters array from main routine
        color Ch: channel name for making color mask image
        stackSize: Color file or MonoChrome / 1 is Mono 3 is Color
        sequence: Mask sequcence No.
    """
#  input Parameters
    if parm is not None: # if parm is not null, then get paraneters...
        Mode, YTh, blInvert, zoneLTh, zoneHTh, TrFac, Show, Gaussian, GaussianSigma,\
            MaskInvert, NextMask = parm 
        """
        variables
           Mode: Image Format 16-bit or 8-bit
           YTh: yellow(or blue) value threshold of mask
           zoneLTh: lower threshold of mask
           zoneHTh: upper threshold of mask
           blinvert: make primary color (blue) mask (negative mask in Luminosity) or not
           Show: working wndows Showing Mode
           Gaussian: Apply or not Gaussian Blur
           GaussianSigma: Gaussian Blur Sigma
           MaskInvert: Mask Inversion for excluding selected area (True / False)
           NextMask: Making Next Mask
        """
        imp0.show()
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
    if stackSize!=1: #if imp0 is color then...
## R ch.
        impR =  impsCh[0]
## G ch.
        impG =  impsCh[1]
## B ch. & B x 0.3
        impB =  impsCh[2]
# Show Windows
#    imp0.close()
    if Show == "Yes":
        if stackSize!=1: #if imp0 is color then...
            impR.show()
            impG.show()
            impB.show()
        imp0.hide()
    if colorCh == "Yellow/Blue":
        imp1 = impB
        imp2 = impG
        imp3 = impR
    elif colorCh == "Cyan/Red":
        imp1 = impR
        imp2 = impG
        imp3 = impB
    elif colorCh == "Magenta/Green":
        imp1 = impG
        imp2 = impR
        imp3 = impB
    elif colorCh == "Luminosity":
        imp1 = Duplicator().run(imp0)
        IJ.run(imp1, "Conversions...", "weighted")
        if Mode == "16-bit":
            ImageConverter(imp1).convertToGray16()
        elif Mode == "8-bit":
            ImageConverter(imp1).convertToGray8()
        imp1.getProcessor().setLut(None)

# Grayscale Mask making process start
    if colorCh != "Luminosity": # not Luminosity
#$$ Make relative RGB Color Mask
        imp2MixRatio = 0.5 # define imp2 Mixing Ratio = 0.5 (2022.5)
        impYmask, V = grayscale_mask_maker(imp1, imp2, imp3, YTh, Mode, blInvert, \
            Show, colorCh, TrFac, Gaussian, GaussianSigma, imp2MixRatio)
#    save relative RGB color Mask
#     Mask inversion (add 0.15)
        if MaskInvert == False or sequence ==2: # Mask No Inversion
            IJ.saveAsTiff(impYmask, folder + filename + "_Mask_Thr_".decode('utf-8') \
                + V.decode('utf-8') + "_x4x" + str(TrFac) + ". tif".decode('utf-8')) 
                    # Save the Subtracted image x 4
        elif sequence ==1: # Mask Inversion for 1st mask
            IJ.run(impYmask, "Invert", "")
            IJ.saveAsTiff(impYmask, folder + filename + "_Mask_Thr_".decode('utf-8') \
                + V.decode('utf-8') + "_x4x" + str(TrFac) + "_Inv. tif".decode('utf-8'))
                    # Save the Subtracted image x 4
    else:
#$$ Make Luminosity Mask
        impYmask, V, V1 = luminosity_mask_maker(imp1, zoneLTh, Mode, blInvert, \
            Show, TrFac)
#    save Luminosity Mask
#     Mask inversion (add 0.15)
        if MaskInvert == False or sequence ==2: # Mask No Inversion
            IJ.saveAsTiff(impYmask, folder + filename + "_Mask_Thr_".decode('utf-8') \
                + V1.decode('utf-8') + "_x4x" + str(TrFac) + ". tif".decode('utf-8')) \
                    # Save the Subtracted image x 4 x transparency factor
        elif sequence == 1: # Mask Inversion for 1st mask
            IJ.run(impYmask, "Invert", "")
            IJ.saveAsTiff(impYmask, folder + filename + "_Mask_Thr_".decode('utf-8') \
                + V1.decode('utf-8') + "_x4x" + str(TrFac) + "_Inv. tif".decode('utf-8'))
                    # Save the Subtracted image x 4 x transparency factor
# Binary Mask Making process start
# select target channel for making lightness range binary mask
#     impZone: base channel image 
    if baseCh == "Luminosity":
        impZone = Duplicator().run(imp0)
        IJ.run(impZone, "Conversions...", "weighted")
        chName ="Gray"
        ImageConverter(impZone).convertToRGB()
    elif baseCh == "Blue":
        impZone = Duplicator().run(impB)
        chName ="B"
    elif baseCh == "Red":
        impZone = Duplicator().run(impR)
        chName ="R"
    elif baseCh == "Green":
        impZone = Duplicator().run(impG)
        chName ="G"
    if stackSize !=1:
        impR.close()
        impG.close()
        impB.close()
    imp0.close()
#$$ Make binary mask for deciding mask valid lightness range
    maskimp = make_binary_mask(impZone, zoneLTh, zoneHTh, Mode, Show)
#      maskimp may be 8bit or 16bit depending on original image bit
#   after deepen bit depth, Save Zone Mask
    IJ.saveAsTiff(maskimp, folder + filename +"_" + chName + "_Zone" + str(int(zoneLTh))\
        + "-" + str(int(zoneHTh)) + "_Mask.tif")
#--------------------------------------------------
#$$ bind valid lightness range binary mask and relative RGB color mask
    ic = plugin.ImageCalculator()
    maskBindimp = ic.run("Min created", impYmask, maskimp) # bind masks
    if Show == "Yes":
        maskBindimp.show()
#   apply Gaussian blur
    if (Gaussian == True) and (colorCh !="Luminosity"):
        IJ.run(maskBindimp, "Gaussian Blur...", "sigma=" + str(GaussianSigma))
#   Mask Inversion for sequence 2
    if MaskInvert == True and sequence ==2:
        IJ.run(maskBindimp, "Invert", "")
##### save binded mask
    #IJ.showMessage("sequence: " + str(sequence) + "/ NextMask: " + str(NextMask))
    if sequence == 2 or NextMask == True: # saving file name for Binding 2 Masks
        IJ.saveAsTiff(maskBindimp, folder + filename + "_tmpMask_".decode('utf-8') \
            + str(sequence).decode('utf-8') + ". tif".decode('utf-8')) 
                # Save the Subtracted image x 4
    else: # saving file name for 1 Mask only
        if MaskInvert == False: # Mask No Inversion
            IJ.saveAsTiff(maskBindimp, folder + filename +"_" + chName + "_Zone" + \
              str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + "_" + V + "_Mask_Binded.tif")
        else: # Mask Inversion
            IJ.saveAsTiff(maskBindimp, folder + filename +"_" + chName + "_Zone" + \
                str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + "_" + V + \
                "_Mask_Binded_and_Inverted.tif")
#     close working windows
    if Show == "Yes":
        impYmask.hide()
        impZone.hide()
        maskimp.hide()
        maskBindimp.hide()
    impYmask.close()
    maskimp.close()
    maskBindimp.close()
    impZone.close()
#   End of def RGBmanipulation 
#----------------------------------------------
class ThresholdPreviewer(AdjustmentListener, ItemListener): 
         # previewer for binary mask and observe changes of dialog items
    def __init__(self, imp, slider_upper, slider_lower, checkbox_prv):
        """
             imp: an ImagePlus
             slider: a java.awt.Scrollbar UI element
             preview_checkbox: a java.awt.Checkbox controlling whether to
                                                 dynamically update the ImagePlus as the
                                                 scrollbar is updated, or not.
        """
        self.imp = imp
        self.imp.setTitle("valid lightness range preview")
        self.original_ip = imp.getProcessor().duplicate() # store a copy
        self.imp_org = Duplicator().run(imp0)
        self.slider_upper = slider_upper
        self.slider_lower = slider_lower
        self.checkbox_prv = checkbox_prv

    def adjustmentValueChanged(self, event):
        """ event: an AdjustmentEvent with data on the state of the scroll bar. """
        preview = self.checkbox_prv.getState()
        if preview:

            if (IJ.isMacOSX() == False) and event.getValueIsAdjusting():
                return 
                  # there are more scrollbar adjustment events queued already
            self.threshold()

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        if event.getStateChange() == event.SELECTED:
            self.threshold()
    
    def reset(self):
        """ Restore the original ImageProcessor """
        self.imp.setProcessor(self.original_ip)
    
    def threshold(self):
        """ Execute the in-place making binary image of the ImagePlus. """
        ip = self.imp.getProcessor()
        mask = ip.createProcessor(ip.width, ip.height)
        mask.BLACK_AND_WHITE_LUT
        mask.setValue(convValue(255, Mode))
        threshold_upper = self.slider_upper.getValue()
        threshold_lower = self.slider_lower.getValue()
        if threshold_upper < threshold_lower:
            threshold_upper = threshold_lower
        self.original_ip.setThreshold(threshold_lower, threshold_upper, \
            ImageProcessor.OVER_UNDER_LUT)
        roi = ThresholdToSelection().convert(self.original_ip) 
            # make ROI from threshold
        mask.setRoi(roi)
        mask.fill(roi)
        self.imp.setProcessor(mask)
        self.imp.getProcessor().setLut (None)
        IJ.selectWindow(self.imp.getTitle())

#----------------------------------------------
class ThresholdPreviewer2(ItemListener, TextListener): 
                  # Previwer for RGB Color mask and observe changes of dialog items
    def __init__(self, impPrv, impPrvR, impPrvG, impPrvB, C0, C1, nfield_maskThr, \
        nfield_TrFac, checkbox_prv, checkbox_Bmask, checkbox_inverse, \
                                                            nfield_slider_lower):
        """
             imp: an ImagePlus
             slider: a java.awt.Scrollbar UI element
             preview_checkbox: a java.awt.Checkbox controlling whether to
                                    dynamically update the ImagePlus as the
                                                 scrollbar is updated, or not.
             nfield: a java.awt.TextField
        """
        self.impPrv = impPrv
        self.impPrv.setTitle("relative color mask preview")
        self.impPrvR = impPrvR
        self.impPrvG = impPrvG
        self.impPrvB = impPrvB
        self.C0 = C0
        self.C1 = C1
        self.C = C0
        self.checkbox_prv = checkbox_prv
        self.checkbox_Bmask = checkbox_Bmask
        self.checkbox_inverse = checkbox_inverse
        self.nfield_maskThr = nfield_maskThr
        self.nfield_TrFac = nfield_TrFac
        self.nfield_slider_lower = nfield_slider_lower
        textStrings = self.nfield_maskThr.getText() #  field for color threshold
        textStrings2 = self.nfield_TrFac.getText() 
                             #  field for transparency factor for mask
        textStrings3 = self.nfield_slider_lower.getText() 
                             #  field for transparency factor for mask
        self.maskThr = int(textStrings) # color threshold for mask
        self.maskTrFac =  float(textStrings2) # transparency factor for mask
        self.Inverse = False # Initialize self.Inverse > invert the image?
        self.LTh = int(textStrings3)

        if C0 == "Yellow":
            self.impPrv0 = Duplicator().run(impPrvB)
            impPrv1 = Duplicator().run(impPrvR)
            impPrv2 = Duplicator().run(impPrvG)
        elif C0 == "Cyan":
            self.impPrv0 = Duplicator().run(impPrvR)
            impPrv1 = Duplicator().run(impPrvG)
            impPrv2 = Duplicator().run(impPrvB)
        elif C0 == "Magenta":
            self.impPrv0 = Duplicator().run(impPrvG)
            impPrv1 = Duplicator().run(impPrvR)
            impPrv2 = Duplicator().run(impPrvB)
        else: # Luminosity
            self.impPrv0 = Duplicator().run(impPrv)
              # Luminosity also has preview image
        self.original_ip = self.impPrv0.getProcessor().duplicate() 
        self.imp_org = Duplicator().run(self.impPrv0)
                                                    # store a copy
        if self.C0 != "Luminosity Positive":
            IJ.run(impPrv1, "Multiply...", "value=0.5")
            IJ.run(impPrv2, "Multiply...", "value=0.5")
            ic = plugin.ImageCalculator()
            self.impPrvTmp = ic.run("Add created", impPrv1, impPrv2)


    def textValueChanged(self, event):
        """ event: a textValueChanged with data on the state of the numeric field. """
        try:
            textStrings = self.nfield_maskThr.getText()
            textStrings2 = self.nfield_TrFac.getText()
            textStrings3 = self.nfield_slider_lower.getText()
            self.maskThr = int(textStrings) # threshold for mask
            self.maskTrFac = float(textStrings2) # Transparency factor for mask
            self.LTh = int(textStrings3) # Transparency factor for mask

        except:
            return
        preview = self.checkbox_prv.getState()
        if preview:
            if self.C0 != "Luminosity Positive":
                self.threshold()
            else:  #Luminosity Mask
                self.luminosity_prv()
        return

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        #if event.getStateChange() == event.SELECTED:
        Bmask = self.checkbox_Bmask.getState()
                # Bmask value: True/False
        if Bmask == True:
            self.C = self.C1
        else:
            self.C = self.C0
        self.Inverse = self.checkbox_inverse.getState()
        preview = self.checkbox_prv.getState()
        if preview:
            if self.C0 != "Luminosity Positive":
                self.threshold()
            else:
                self.luminosity_prv()
    
    def reset(self):
        """ Restore the original ImageProcessor """
        self.imp.setProcessor(self.original_ip)
    
    def threshold(self):
        ic = plugin.ImageCalculator()

        """ Execute the in-place making binary image of the ImagePlus. """

        if self.C == self.C1: # for primary mask
            self.maskThr2 = self.maskThr * -1 
                   # Invert the threshold value for primary mask
        else:  # for complimentary mask
            self.maskThr2 = self.maskThr
               # not invert the threshold value for complimentary mask
        V2 = str(self.maskThr2) # actual Yellow mask Threshold
            # for making gray scale mask
               # Set the Threshold in R + G Image
        impPrvTmp = Duplicator().run(self.impPrvTmp)
        IJ.run(impPrvTmp, "Add...","value=" + V2)

        if self.C == self.C0: # Ymask
            impPrvMask = ic.run("Subtract created", impPrvTmp, self.impPrv0)
        else: # Bmask (Invert)
            impPrvMask = ic.run("Subtract created", self.impPrv0, impPrvTmp)
        mask_factor = 4 * self.maskTrFac
        IJ.run(impPrvMask, "Multiply...","value=" + str(mask_factor))
        if self.Inverse == True:
            IJ.run(impPrvMask, "Invert","")
        ip_PrvMask = impPrvMask.getProcessor()
        ip_PrvMask.BLACK_AND_WHITE_LUT
        self.impPrv.setProcessor(ip_PrvMask)
        self.impPrv.getProcessor().setLut (None)
        impPrvTmp.close()
        impPrvMask.close()
        IJ.selectWindow(self.impPrv.getTitle())

    def luminosity_prv(self):
    # Make Grayscale Mask Image Files 
    # Target (Yellow) Mask Threshold
        V = str(self.LTh) # for threshold expression in file name
        impPrvTmp = Duplicator().run(self.impPrv0)
        IJ.run(impPrvTmp, "Subtract...","value=" + V)
#        IJ.showMessage("Luminosity self.C: " + self.C)
        if self.C == self.C1: # for making inverted Negative mask
            IJ.run(impPrvTmp, "Multiply...", "value=" + str(float(4 / self.maskTrFac))) 
                # making mask more transparent (x 4 for Nega)
            impPrvTmp.getProcessor().invert()
        else: # for making positive mask
            IJ.run(impPrvTmp, "Multiply...", "value=" + str(float(4 * self.maskTrFac))) 
                # making mask more transparent (x 4)
        ip_PrvMask = impPrvTmp.getProcessor()
        ip_PrvMask.BLACK_AND_WHITE_LUT
        self.impPrv.setProcessor(ip_PrvMask)
        self.impPrv.getProcessor().setLut (None)
        impPrvTmp.close()
        IJ.selectWindow(self.impPrv.getTitle())

#-------------------------------------------
class LuminosityMaskPreviewer(AdjustmentListener, ItemListener, TextListener): 
         # previewer for binary mask and observe changes of dialog items
    def __init__(self, impPrv, slider_upper, slider_lower, C0, C1, \
        nfield_TrFac, checkbox_prv, checkbox_Bmask, checkbox_inverse):
        """
             imp: an ImagePlus
             slider: a java.awt.Scrollbar UI element
             preview_checkbox: a java.awt.Checkbox controlling whether to
                                                 dynamically update the ImagePlus as the
                                                 scrollbar is updated, or not.
        """

        self.impPrv = impPrv
        self.impPrv.setTitle("luminosity mask preview")
        self.original_ip = impPrv.getProcessor().duplicate() # store a copy
        self.slider_upper = slider_upper
        self.slider_lower = slider_lower
        self.C0 = C0
        self.C1 = C1
        self.C = C0
        self.checkbox_prv = checkbox_prv # Preview checkbox
        self.checkbox_Bmask = checkbox_Bmask # Negative / Positive checkbox
        self.checkbox_inverse = checkbox_inverse 
              #Attention! use it for exchanging preview image in Luminosity mask 
        self.nfield_TrFac = nfield_TrFac # transparency factor
        textStrings2 = self.nfield_TrFac.getText() 
                             #  field for transparency factor for mask
        self.maskTrFac =  float(textStrings2) # transparency factor for mask
        self.Inverse = False # False: grayscale mask / True: binary mask

        if C0 == "Luminosity Positive":
            self.impPrv0 = Duplicator().run(impPrv)
              # Luminosity also has preview image
        self.original_ip = self.impPrv0.getProcessor().duplicate() 
        self.imp_org = Duplicator().run(self.impPrv0)
                                                    # store a copy


    def adjustmentValueChanged(self, event):
        """ event: an AdjustmentEvent with data on the state of the scroll bar. """
        preview = self.checkbox_prv.getState()
        if preview:

            if (IJ.isMacOSX() == False) and event.getValueIsAdjusting():
                return 
                  # there are more scrollbar adjustment events queued already
            #self.threshold()
            self.BinaryPrv = self.checkbox_inverse.getState()
            if self.BinaryPrv == True:
                self.threshold()
            else:
                self.luminosity_prv()

    def textValueChanged(self, event):
        """ event: a textValueChanged with data on the state of the numeric field. """
        try:
            textStrings2 = self.nfield_TrFac.getText()
            self.maskTrFac = float(textStrings2) # Transparency factor for mask

        except:
            return
        preview = self.checkbox_prv.getState()
        if preview:

            self.BinaryPrv = self.checkbox_inverse.getState()
            if self.BinaryPrv == True:
                self.threshold()
            else:
                self.luminosity_prv()
        return

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        Bmask = self.checkbox_Bmask.getState()
        #IJ.showMessage("Bmask state: " + str(Bmask))
                # Bmask value: True/False
        if Bmask == True:
            self.C = self.C1
        else:
            self.C = self.C0
        self.BinaryPrv = self.checkbox_inverse.getState()
        if self.BinaryPrv == True:
            self.threshold()
        else:
            self.luminosity_prv()
    
    def reset(self):
        """ Restore the original ImageProcessor """
        self.impPrv.setProcessor(self.original_ip)
    
    def threshold(self):
        """ Execute the in-place making binary image of the ImagePlus. """
        ip = self.impPrv.getProcessor()
        mask = ip.createProcessor(ip.width, ip.height)
        mask.BLACK_AND_WHITE_LUT
        mask.setValue(convValue(255, Mode))
        threshold_upper = self.slider_upper.getValue()
        threshold_lower = self.slider_lower.getValue()
        if threshold_upper < threshold_lower:
            threshold_upper = threshold_lower
        self.original_ip.setThreshold(threshold_lower, threshold_upper, \
            ImageProcessor.OVER_UNDER_LUT)
        roi = ThresholdToSelection().convert(self.original_ip) 
            # make ROI from threshold
        mask.setRoi(roi)
        mask.fill(roi)
        self.impPrv.setProcessor(mask)
        self.impPrv.getProcessor().setLut (None)
        IJ.selectWindow(self.impPrv.getTitle())

    def luminosity_prv(self):
    # Make Grayscale Mask Image Files 
    # Target (Yellow) Mask Threshold
        #threshold_upper = self.slider_upper.getValue()
        threshold_lower = self.slider_lower.getValue()
        V = str(threshold_lower) # for threshold expression in file name
        impPrvTmp = Duplicator().run(self.impPrv0)
        IJ.run(impPrvTmp, "Subtract...","value=" + V)
#        IJ.showMessage("Luminosity self.C: " + self.C)
        if self.C == self.C1: # for making inverted Negative mask
            IJ.run(impPrvTmp, "Multiply...", "value=" + str(float(4 / self.maskTrFac))) 
                # making mask more transparent (x 4 for Nega)
            impPrvTmp.getProcessor().invert()
        else: # for making positive mask
            IJ.run(impPrvTmp, "Multiply...", "value=" + str(float(4 * self.maskTrFac))) 
                # making mask more transparent (x 4)
        ip_PrvMask = impPrvTmp.getProcessor()
        ip_PrvMask.BLACK_AND_WHITE_LUT
        self.impPrv.setProcessor(ip_PrvMask)
        self.impPrv.getProcessor().setLut (None)
        impPrvTmp.close()
        IJ.selectWindow(self.impPrv.getTitle())



#---------------------------------------------

def thresholdImageUI(imp, impPrvR, impPrvG, impPrvB, Mode, C0, C1, sequence):
    ImageConverter(imp).convertToGray8()
    WM.getActiveWindow().toFront()
    gd = GenericDialog(C0 +" (" + C1 + \
                                  ") Mask Maker / input threshold parameters")
    if C0 != "Luminosity Positive":
        gd.addNumericField(C0 +" (" + C1 + ") Mask Threshold \n" + \
                                               " (Signed relative value)", 10, 0)
    else:
        gd.addNumericField \
                       ("This field is not available for Luminosity Mask", 0, 0)
    if C0 != "Luminosity Positive":
        gd.addCheckbox("Make " + C1 + " (primary color) Mask", False)
    else:
        gd.addCheckbox("Make " + C1 + " Mask", False)
    gd.addMessage("Decide mask valid lightness range (0-255)\n"+ \
                        "         White part in preview is mask valid region")
    gd.addSlider("Upper Threshold of Mask", 0, 255, 128)
    gd.addSlider("Lower Threshold of Mask", 0, 255, 128)
    gd.addCheckbox("Preview", True)
    gd.addNumericField("Mask Transparency Factor", 1.0, 1)
    if C0 != "Luminosity Positive":
        gd.addCheckbox("Apply Gaussian Blur", False)
        gd.addNumericField("Gaussian Blur Sigma", 2.0, 1)
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])
    if C0 != "Luminosity Positive":
        gd.addMessage("to make binary mask instead of grayscale mask, \n"\
            + "  set Mask Transparency Factor = 100")
    if C0 != "Luminosity Positive":
        gd.addCheckbox("Mask Inversion for Excluding selected area", False)
    else:
        gd.addCheckbox("Show Lightness range preview", False)

    if sequence ==1:
        gd.addCheckbox("Make combined Mask", False)


    # The UI elements for the above inputs
    slider_upper = gd.getSliders().get(0) # the 1st slider
    slider_lower = gd.getSliders().get(1) # the 2nd slider
    checkbox_Bmask = gd.getCheckboxes().get(0) 
        #  chkbox for primary color mask / or negative mask (Luminosity)
    checkbox_prv = gd.getCheckboxes().get(1) # chkbox for preview
    nfield_maskThr = gd.getNumericFields().get(0) # 1st NumericField
    nfield_slider_lower = gd.getNumericFields().get(2) # 2nd slider's NumericField
       # Attention! sliders also have NumericFields 
    nfield_TrFac = gd.getNumericFields().get(3) # Transparency Factor
    choice_showWin = gd.getChoices().get(0)
    if C0 != "Luminosity Positive": # RGB color mask
        checkbox_inverse = gd.getCheckboxes().get(3)
    else:     # in Luminosity mask checkbox_inverse: change preview gray/bin. 
        checkbox_inverse = gd.getCheckboxes().get(2) # Attention index number!
        """
         in Luminosity mask,
           checkbox_inverse & checkbox_Bmask have same function (make negative mask)
        """
          # checkbox _inverse < no use for Luminosity grayscale mask


    if not imp:
        print "Open an image first!"
        return
    # add event listeners
    #--add previewer1--- for binary mask
    impcp = Duplicator().run(imp)
    if C0 != "Luminosity Positive":
        previewer = ThresholdPreviewer(imp, slider_upper, slider_lower, checkbox_prv)
        slider_upper.addAdjustmentListener(previewer) # set listener for upper slider
        slider_lower.addAdjustmentListener(previewer) # set listener for lower slider
        checkbox_prv.addItemListener(previewer)  # set listener for preview checkbox
    #--add previewer2--- for grayscale mask
    impcp.show()
    if C0 != "Luminosity Positive":
        previewer2 = ThresholdPreviewer2(impcp, impPrvR, impPrvG, impPrvB, C0, C1, \
        nfield_maskThr, nfield_TrFac, checkbox_prv, checkbox_Bmask, \
        checkbox_inverse, nfield_slider_lower)
        checkbox_Bmask.addItemListener(previewer2)
        checkbox_prv.addItemListener(previewer2)
        nfield_maskThr.addTextListener(previewer2)
        nfield_TrFac.addTextListener(previewer2)
        checkbox_inverse.addItemListener(previewer2)
    #-- add Luminosity Mask Previewer--
    if C0 == "Luminosity Positive":
        previewerL = LuminosityMaskPreviewer(impcp, slider_upper, slider_lower,\
            C0, C1, nfield_TrFac, checkbox_prv, checkbox_Bmask, checkbox_inverse)
        slider_upper.addAdjustmentListener(previewerL) # set listener for upper slider
        slider_lower.addAdjustmentListener(previewerL) # set listener for lower slider
        checkbox_Bmask.addItemListener(previewerL)
        checkbox_prv.addItemListener(previewerL)  # set listener for preview checkbox
        nfield_TrFac.addTextListener(previewerL)  
            # set listener for transparency fac. checkbox
        checkbox_inverse.addItemListener(previewerL) 
            # in here use it for exchanging preview image

#   show daialog
    gd.showDialog()

    if gd.wasCanceled():
        if C0 != "Luminosity Positive":
            previewer.reset()
            previewer2.reset()
        else:
            previewerL.reset()
        Mode = "Stop"
        print "User canceled dialog!"
#   input parameters 
    YTh =  gd.getNextNumber()
    blInvert = gd.getNextBoolean()
    zoneHTh =  gd.getNextNumber()
    zoneLTh =  gd.getNextNumber()
    preview_on = gd.getNextBoolean()
    TrFac =  gd.getNextNumber()
    if C0 != "Luminosity Positive":
        Gaussian = gd.getNextBoolean()
        GaussianSigma =  gd.getNextNumber()
    else:
        Gaussian = False
        GaussianSigma =  0.0
    Show =  gd.getNextChoice()
    if C0 != "Luminosity Positive": # RGB color mask
        MaskInvert =  gd.getNextBoolean()
    else: # Luminosity mask
        MaskInvert =  False
    if sequence ==1: # 1st dialog
        NextMask =  gd.getNextBoolean()
    else: # 2nd dialog
        NextMask =  False
    # all preview images will be closed
    #--binaly preview image
    imp.hide()
    imp.close()
    #--grayscale preview image (rgb color & luminosity)
    impcp.hide()
    impcp.close()
    # remove event listeners
    #--remove previewer1--- for binary mask
    if C0 != "Luminosity Positive":
        slider_upper.removeAdjustmentListener(previewer) # set listener for upper slider
        slider_lower.removeAdjustmentListener(previewer) # set listener for lower slider
        checkbox_prv.removeItemListener(previewer)  # set listener for preview checkbox
    #--remove previewer2--- for grayscale mask
    if C0 != "Luminosity Positive":
        checkbox_Bmask.removeItemListener(previewer2)
        checkbox_prv.removeItemListener(previewer2)
        nfield_maskThr.removeTextListener(previewer2)
        nfield_TrFac.removeTextListener(previewer2)
        checkbox_inverse.removeItemListener(previewer2)
    #--remove previewerL--- for Luminosity mask
    if C0 == "Luminosity Positive":
        slider_upper.removeAdjustmentListener(previewerL) 
            # remove listener for upper slider
        slider_lower.removeAdjustmentListener(previewerL)
            # remove listener for lower slider
        checkbox_prv.removeItemListener(previewerL)
            # remove listener for preview checkbox
        nfield_TrFac.removeTextListener(previewerL)  
            # remove listener for transparency fac. checkbox
        checkbox_inverse.removeItemListener(previewerL) 
            # in here use it for exchanging preview image
    return Mode, YTh, blInvert, zoneLTh, zoneHTh, TrFac, str(Show), Gaussian, \
        GaussianSigma, MaskInvert, NextMask
    """
    return values
        YTh: Target Color Ch Threshold (originally Yellow Threshold)
        zoneLTh: lower threshold of mask
        zoneHTh: upper threshold of mask
        blInvert: Color invert Yes or No (originally invert for blue mask) 
        Show: working wndows Showing Mode
        TrFac: Transparency Factor
        Gaussian: Apply or not Gaussian Blur
        GaussianSigma: Gaussian Blur Sigma
        MaskInvert: Mask Inversion for excluding selected area (True / False)
    """

#-------------------------------------------
class DialogPreviewer(ItemListener):    
    def __init__(self, choice1, choice2):    
        """ 
             choice: a java.awt.choice
        """    
        self.choice1 = choice1    
        self.choice2 = choice2    

 
    def itemStateChanged(self, event):    
        """ event: an ItemEvent with data on what happened to the checkbox. """    
        choice_item  = self.choice1.getSelectedItem()
        #print "Dialog Previewr: ", str(choice_item)
        if choice_item == "Luminosity":
            choice_item2 = "Luminosity"
        elif choice_item == "Yellow/Blue":
            choice_item2 = "Blue"
        elif choice_item == "Cyan/Red":
            choice_item2 = "Red"
        else: # Magenta/Green
            choice_item2 = "Green"
        self.choice2.select(choice_item2)    
        return

#-----------------------------
def selectChannel(): 
                 # dialog of selecting target channel for mask valid lightness range
    gd = GenericDialog("General Color Channel Mask Maker / Target Channel Selection")
    scolorCh = ["Luminosity", "Cyan/Red","Magenta/Green","Yellow/Blue"]
    gd.addChoice("Select target channel for making mask", scolorCh, scolorCh[2])
    sbaseCh = ["Luminosity","Red","Green","Blue"]
    gd.addChoice("Select target channel for deciding mask valid lightness range", \
                                                                   sbaseCh, sbaseCh[2])
    gd.addMessage("If you select Luminosity as making mask ch., \n"\
            + "  lightness range deciding ch. will be coerced to Luminosity")

    # The UI elements for the above two inputs    
    choice1 = gd.getChoices().get(0)
    choice2 = gd.getChoices().get(1)
    
    previewer = DialogPreviewer(choice1, choice2)    
    choice1.addItemListener(previewer)    

    gd.showDialog()

    colorCh =  gd.getNextChoice()
    baseCh =  gd.getNextChoice()
    if gd.wasCanceled():
        colorCh =  "None"
        baseCh =  "None"

    if colorCh == "Luminosity" and baseCh != "Luminosity":
        baseCh = "Luminosity"
        IJ.showMessage("lightness range deciding ch. is changed to Luminosity")
    return str(baseCh),  str(colorCh)
    """
    return values
        baseCh: channel name for deciding mask upper & lower thresholds
        color Ch: channel name for making color mask image
    """

#-------------------------------------------
def parameter_input(stackSize, sequence):
## opening parameter input dialog 1
#   parameter input for deciding base channel and color mask channel 
    sequence = sequence + 1
    if stackSize == 1: # if imp0 is Monochrome
        baseCh = "Luminosity"
        colorCh = "Luminosity"
    else: # if imp0 is RGB Color
        baseCh, colorCh = selectChannel() 
                           # select target channel for deciding upper & lower threshold
        if baseCh == "None":
            imp0.hide()
            imp0.close()
    if colorCh == "Yellow/Blue":
        C0 = "Yellow"
        C1 = "Blue"
    elif colorCh == "Cyan/Red":
        C0 = "Cyan"
        C1 = "Red"
    elif colorCh == "Magenta/Green":
        C0 = "Magenta"
        C1 = "Green"
    elif colorCh == "Luminosity":
        C0 = "Luminosity Positive"
        C1 = "Negative"
    if baseCh != "None":
        imp0.hide()

        if baseCh == "Blue":
            impPrvBase = Duplicator().run(impprvsCh[2])
        elif baseCh == "Red":
            impPrvBase = Duplicator().run(impprvsCh[0])
        elif baseCh == "Green":
            impPrvBase = Duplicator().run(impprvsCh[1])
        else:  # Luminosity
            ImageConverter(tmpimpprv).convertToGray8()
            impPrvBase = Duplicator().run(tmpimpprv)
        impPrvBase.show()
        parm = thresholdImageUI(impPrvBase, impprvsCh[0], impprvsCh[1], \
                                        impprvsCh[2], Mode, C0, C1, sequence) 

## opening parameter input dialog 2
#    threshold parameters input for making masks
#          (for valid ligheness range or color threshold)
    else: # if parameter input dialog1 was canceled
        parm = []
        parm.append("Stop")
    return parm, baseCh, colorCh, stackSize, sequence

#------------ main routine -----------#
window = WM.getActiveWindow()
Mode = "Auto"
if not window: # if no opening image > open file
    """ opening file """
    od = OpenDialog("Choose a file")
    folder = od.getDirectory()
    filename = od.getFileName()
    if filename:
        path = folder + filename
        IJ.log(path)
        IJ.log(folder)
        IJ.log(filename)
        # Always Open file with Bio-format Plugin
        imps = BF.openImagePlus(folder + filename) # Open file with Bio-format Plugin
        imp0 = imps[0] #imp0: Original Image 
        imp0.show()
    else:
        Mode = "Stop"
else: # target image is already open
    imp0 = IJ.getImage()
    filename = imp0.getTitle()
    try:
        folder = imp0.getOriginalFileInfo().directory
    except:
        folder = IJ.getDirectory("Output_directory")
# deciding bit depth
if Mode == "Auto":
    bitdepth = imp0.getBitDepth()
    if bitdepth == 8:
            Mode = "8-bit"
    elif bitdepth ==16:
            Mode = "16-bit"
    elif bitdepth ==32:
            Mode = "32-bit"
    else:
            Mode = "16-bit"
stackSize =imp0.getStackSize() # Color or Monochrome?
if Mode != "Stop": # if user did not cancel executing process
    """ making preview image """
    impprv = Duplicator().run(imp0) # impprv: preview imp
    scale = 1200.0 / impprv.getWidth() 
                        # preview image scaling value (width = 1200px fixed)
    new_height = impprv.getHeight() * scale # preview image height
    new_height = int(new_height)
    tmpimpprv = impprv.resize(1200, new_height, "none")
    ImageConverter(tmpimpprv).convertToRGB() # convert to 8bit RGB
    impprvsCh = ChannelSplitter.split(tmpimpprv)
    """ start for defining parameters and making masks process """
    sequence = 0  # 1st mask image
    parm, baseCh, colorCh, stackSize, sequence = parameter_input(stackSize, sequence)
    if parm[0] != "Stop": # if parameter input dialog2 was not canceled
        """ proceeding 1st mask making process """
        RGBmanipulation(imp0, folder, filename, baseCh, parm, colorCh, stackSize, \
                                                                             sequence)
        if parm[10] == True:  # Making 2nd mask image
            parm, baseCh, colorCh, stackSize, sequence = parameter_input(stackSize,\
                                                                             sequence)
            if parm[0] != "Stop": # if parameter input dialog2 was not canceled
                """ proceeding 2nd mask making process """
                imps100 = BF.openImagePlus(folder + filename + "_tmpMask_1.tif") 
                RGBmanipulation(imp0, folder, filename, baseCh, parm, colorCh, \
                                                                  stackSize, sequence)
                """ combining 2 mask images """
                if sequence ==2: # if there are 2 masks then combine 2 masks
                    imps101 = BF.openImagePlus(folder + filename + "_tmpMask_2.tif") 
                    ic = plugin.ImageCalculator()
                    maskLastBindimp = ic.run("Min created", imps100[0], imps101[0])
                                                                        # bind masks
                    IJ.saveAsTiff(maskLastBindimp, folder + filename + \
                                                                "_combined_mask.tif")
#
    # closing temporary windows for preview images
    tmpimpprv.hide()
    impprvsCh[0].hide()
    impprvsCh[1].hide()
    impprvsCh[2].hide()
    tmpimpprv.close()
    impprvsCh[0].close()
    impprvsCh[1].close()
    impprvsCh[2].close()
    
else: # if user canceled file opening
    IJ.log("User canceled")
IJ.log("Finish")

